"""Mixins for different device functionalities"""
import decimal

from ..scpi import SCPIDevice


class MultiMeter(SCPIDevice):
    """Multimeter related features"""

    async def measure_voltage(self, extra_params: str = "") -> decimal.Decimal:
        """Returns the measured (scalar) actual output voltage (in volts),
        pass extra_params string to append to the command (like ":ACDC")"""
        resp = await self.ask(f"MEAS:SCAL:VOLT{extra_params}?")
        return decimal.Decimal(resp)

    async def measure_current(self, extra_params: str = "") -> decimal.Decimal:
        """Returns the measured (scalar) actual output current (in amps),
        pass extra_params string to append to the command (like ":ACDC")"""
        resp = await self.ask(f"MEAS:SCAL:CURR{extra_params}?")
        return decimal.Decimal(resp)

    async def set_measure_current_max(self, amps: float) -> None:
        """Sets the upper bound (in amps) of current to measure,
        on some devices low-current accuracy can be increased by keeping this low"""
        await self.command(f"SENS:CURR:RANG {amps:f}")

    async def query_measure_current_max(self) -> decimal.Decimal:
        """Returns the upper bound (in amps) of current to measure,
        this is not neccessarily same number as set with set_measure_current_max"""
        resp = await self.ask("SENS:CURR:RANG?")
        return decimal.Decimal(resp)


class PowerSupply(SCPIDevice):
    """Power supply related features"""

    async def set_voltage(self, millivolts: float, extra_params: str = "") -> None:
        """Sets the desired output voltage (but does not auto-enable outputs) in millivolts,
        pass extra_params string to append to the command (like ":PROT")"""
        await self.command(f"SOUR:VOLT{extra_params} {millivolts:f} MV")

    async def query_voltage(self, extra_params: str = "") -> decimal.Decimal:
        """Returns the set output voltage (in volts),
        pass extra_params string to append to the command (like ":PROT")"""
        resp = await self.ask(f"SOUR:VOLT{extra_params}?")
        return decimal.Decimal(resp)

    async def set_current(self, milliamps: float, extra_params: str = "") -> None:
        """Sets the desired output current (but does not auto-enable outputs) in milliamps,
        pass extra_params string to append to the command (like ":TRIG")"""
        await self.command(f"SOUR:CURR{extra_params} {milliamps:f} MA")

    async def query_current(self, extra_params: str = "") -> decimal.Decimal:
        """Returns the set output current (in amps),
        pass extra_params string to append to the command (like ":TRIG")"""
        resp = await self.ask(f"SOUR:CURR{extra_params}?")
        return decimal.Decimal(resp)

    async def set_output(self, state: bool) -> None:
        """Enables/disables output"""
        await self.command(f"OUTP:STAT {state:d}")

    async def query_output(self) -> bool:
        """Returns the output state"""
        resp = await self.ask("OUTP:STAT?")
        return bool(int(resp))



"""HP/Agilent 3362B specific device implementation and helpers"""
from typing import Any
import logging
import decimal

from ..scpi import SCPIDevice, SCPIProtocol
from ..transports.rs232 import RS232Transport
from ..transports.rs232 import get as get_rs232
from .generic import MultiMeter, PowerSupply

LOGGER = logging.getLogger(__name__)


class HP6632B(PowerSupply, MultiMeter, SCPIDevice):
    """Adds the HP/Agilent 3362B specific SCPI commands as methods"""

    async def set_low_current_mode(self, state: bool) -> None:
        """The low-current mode is enabled by setting the range to (max) 20mA, anything over that is high-current mode.
        This model has max 5A output"""
        if state:
            return await self.set_measure_current_max(0.020)
        return await self.set_measure_current_max(5.0)

    async def query_low_current_mode(self) -> bool:
        """Returns boolean indicating whether we are in low or high current mode"""
        max_current = await self.query_measure_current_max()
        if max_current <= 0.020:
            return True
        return False

    async def measure_current_autorange(self, extra_params: str = "") -> decimal.Decimal:
        """Measures the current, then make sure we are running on the correct measurement range
        and if not switch range and measure again"""
        ret = await self.measure_current(extra_params)
        if abs(ret) < 0.020:
            # We need to be in low-current mode
            if not await self.query_low_current_mode():
                # Enter low current mode and measure again
                await self.set_low_current_mode(True)
                return await self.measure_current(extra_params)
            return ret
        # We need to be in high-current mode
        if await self.query_low_current_mode():
            # Switch mode and measure again
            await self.set_low_current_mode(False)
            return await self.measure_current(extra_params)
        return ret

    def ensure_transport_is_rs232(self) -> None:
        """Ensures transport is RS232, raises error if not"""
        if not isinstance(self.protocol.transport, RS232Transport):
            raise RuntimeError("Only usable with RS232 transports")

    async def set_remote_mode(self, state: bool = True) -> None:
        """RS232 only, prevent accidental button mashing on the fron panel, this switches between SYSTem:REMote
        and SYSTem:LOCal according to state, this overrides previous value set with set_rwlock"""
        self.ensure_transport_is_rs232()
        if state:
            return await self.command("SYST:REM")
        return await self.command("SYST:LOC")

    async def set_rwlock(self, state: bool = True) -> None:
        """RS232 only, prevent *any* button mashing on the fron panel, this switches between SYSTem:RWLock
        and SYSTem:LOCal according to state, this overrides previous value set with set_remote_mode"""
        self.ensure_transport_is_rs232()
        if state:
            return await self.command("SYST:RWL")
        return await self.command("SYST:LOC")

    async def display_on(self, state: bool = True) -> None:
        """Sets display on/off"""
        if state:
            return await self.command("DISP ON")
        return await self.command("DISP OFF")

    async def set_display_mode(self, mode: str) -> None:
        """Set the display mode, valied values are NORM and TEXT"""
        mode = mode.upper()
        if mode not in ("NORM", "TEXT"):
            raise ValueError("Invalid mode %s, valid ones are NORM and TEXT" % mode)
        return await self.command(f"DISP:MODE {mode}")

    async def set_display_text(self, text: str) -> None:
        """Used to display text on the display, max 14 characters,
        NOTE: does *not* set display mode, you need to do it yourself"""
        if len(text) > 14:
            raise ValueError("Max text length is 14 characters")
        if '"' in text and "'" in text:
            raise ValueError("Text may only contain either single or double quotes, not both")
        if '"' in text:
            return await self.command(f"DISP:TEXT '{text}'")
        return await self.command(f"""DISP:TEXT  "{text}" """.strip())


def rs232(serial_url: str, **kwargs: Any) -> HP6632B:
    """Quick helper to connect via RS232 port"""
    transport = get_rs232(serial_url, **kwargs)
    protocol = SCPIProtocol(transport)
    dev = HP6632B(protocol)
    return dev



"""Serial port transport layer"""
from __future__ import annotations
from typing import Optional, Any, Dict, cast
import asyncio
import logging
from dataclasses import field, dataclass

import serial  # type: ignore
import serial.threaded  # type: ignore

from .baseclass import BaseTransport

LOGGER = logging.getLogger(__name__)
WRITE_TIMEOUT = 1.0


class RS232SerialProtocol(serial.threaded.LineReader):  # type: ignore
    """PySerial "protocol" class for handling stuff"""

    ENCODING = "ascii"

    def connection_made(self, transport: RS232Transport) -> None:
        """Overridden to make sure we have write_timeout set"""
        super().connection_made(transport)
        # Make sure we have a write timeout of expected size
        self.transport.write_timeout = WRITE_TIMEOUT

    def handle_line(self, line: str) -> None:
        raise RuntimeError("This should have been overloaded by RS232Transport")


@dataclass
class RS232Transport(BaseTransport):
    """Uses PySerials ReaderThread in the background to save us some pain"""

    serialdevice: Optional[serial.SerialBase] = field(default=None)
    _serialhandler: Optional[serial.threaded.ReaderThread] = field(default=None, repr=False)

    def __post_init__(self) -> None:
        """Initialize the transport"""
        if not self.serialdevice:
            raise ValueError("serialdevice must be given")
        self._serialhandler = serial.threaded.ReaderThread(self.serialdevice, RS232SerialProtocol)
        self._serialhandler.start()
        self._serialhandler.protocol.handle_line = self.message_received

    async def send_command(self, command: str) -> None:
        """Wrapper for write_line on the protocol with some sanity checks"""
        if not self._serialhandler or not self._serialhandler.is_alive():
            raise RuntimeError("Serial handler not ready")
        async with self.lock:
            self._serialhandler.protocol.write_line(command)

    async def get_response(self) -> str:
        """Serial devices send responses without needing to be told to, just reads it"""
        # TODO: we probably have a race-condition possibility here, maybe always put all received
        # messages to a stack and return popleft ??
        async with self.lock:
            response: Optional[str] = None

            def set_response(message: str) -> None:
                """Callback for setting the response"""
                nonlocal response, self
                response = message
                self.blevent.set()

            self.blevent.clear()
            self.message_callback = set_response
            await asyncio.get_event_loop().run_in_executor(None, self.blevent.wait)
            self.message_callback = None
            return cast(str, response)

    async def abort_command(self) -> None:
        """Uses the break-command to issue "Device clear", from the SCPI documentation (for HP6632B):
        The status registers, the error queue, and all configuration states are left unchanged when a device
        clear message is received. Device clear performs the following actions:
             - The input and output buffers of the dc source are cleared.
             - The dc source is prepared to accept a new command string."""
        if not self._serialhandler:
            raise RuntimeError("No serialhandler")
        if not self._serialhandler.serial:
            raise RuntimeError("No serialhandler.serial")
        async with self.lock:
            self._serialhandler.serial.send_break()

    async def quit(self) -> None:
        """Closes the port and background threads"""
        if not self._serialhandler:
            raise RuntimeError("No serialhandler")
        if not self._serialhandler.serial:
            raise RuntimeError("No serialhandler.serial")
        self._serialhandler.close()


def get(serial_url: str, **serial_kwargs: Dict[str, Any]) -> RS232Transport:
    """Shorthand for creating the port from url and initializing the transport"""
    port = serial.serial_for_url(serial_url, **serial_kwargs)
    return RS232Transport(serialdevice=port)



#!/usr/bin/env python3
"""Example/test script for use the with Prologix USB GPIB interface"""
import atexit
import os
import sys

from scpi import SCPIDevice
from scpi.transports.gpib import prologix
from scpi.wrapper import AIOWrapper
from scpi.devices.hp6632b import HP6632B
from scpi.devices.generic import MultiMeter

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(f"run with python -i {__file__} /dev/ttyUSB0")
        sys.exit(1)
    # Then put to interactive mode
    os.environ["PYTHONINSPECT"] = "1"
    # Get the low-level GPIB transport
    aiogpib = prologix.get(sys.argv[1])
    # And the mapper that handlers asyncio transparently
    gpib = AIOWrapper(aiogpib)
    atexit.register(gpib.quit)

    print("*** Scanning bus for devices ***")
    devlist = gpib.scan_devices()
    devdict = {}
    for addr, idstr in devlist:
        man, model, _, _ = idstr.split(",")
        # Get device specific transport instance
        dtransport = aiogpib.get_device_transport(addr)
        # Get the device class with the transport
        aiodev = SCPIDevice(dtransport)
        if man == "HEWLETT-PACKARD":
            if model == "6632B":
                aiodev = HP6632B(dtransport)
            if model == "34401A":
                aiodev = MultiMeter(dtransport)
        # And get the mapper that handles asyncio transparently
        devdict[addr] = AIOWrapper(aiodev)
        print("Added {:s} as devdict[{:d}]".format(idstr, addr))



"""TCP based transport"""
from typing import Optional
import asyncio
from dataclasses import dataclass, field
import logging


from .baseclass import BaseTransport


LOGGER = logging.getLogger(__name__)


@dataclass
class TCPTransport(BaseTransport):
    """TCP based transport"""

    ipaddr: Optional[str] = field(default=None)
    port: Optional[int] = field(default=None)
    reader: Optional[asyncio.StreamReader] = field(default=None)
    writer: Optional[asyncio.StreamWriter] = field(default=None)

    async def open_connection(self, ipaddr: str, port: int) -> None:
        """Open a connection (also update the IP/port)"""
        self.reader, self.writer = await asyncio.open_connection(ipaddr, port, loop=asyncio.get_event_loop())
        self.ipaddr = ipaddr
        self.port = port

    def __post_init__(self) -> None:
        """Call open_connection in an eventloop"""
        if self.ipaddr is None or self.port is None:
            raise ValueError("ipaddr and port must be given")
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.open_connection(self.ipaddr, self.port))

    async def send_command(self, command: str) -> None:
        """Write command to the stream"""
        if not self.writer:
            raise RuntimeError("Writer not set")
        async with self.lock:
            LOGGER.debug("sending command: {}".format(command))
            self.writer.write((command + "\r\n").encode())
            await asyncio.sleep(0.05)
            await self.writer.drain()

    async def get_response(self) -> str:
        """Get response from the stream"""
        if not self.reader:
            raise RuntimeError("Reader not set")
        async with self.lock:
            data = await self.reader.readline()
            res = data.decode()
            LOGGER.debug("Got response: {}".format(res.strip()))
            return res

    async def quit(self) -> None:
        """Closes the connection and background threads"""
        if not self.writer:
            raise RuntimeError("Writer not set")
        self.writer.close()
        await self.writer.wait_closed()

    async def abort_command(self) -> None:
        """This does not apply on TCP transport"""
        LOGGER.debug("TCP transport does not know what to do here")


def get(ipaddr: str, port: int) -> TCPTransport:
    """Shorthand for creating the port from ip and port and initializing the transport"""
    return TCPTransport(ipaddr=ipaddr, port=port)



"""GPIB Related baseclasses"""
from __future__ import annotations
from typing import Optional, Tuple, Sequence, Union
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import logging

from ..baseclass import AbstractTransport, BaseTransport

LOGGER = logging.getLogger(__name__)

AddressTuple = Tuple[int, Optional[int]]


@dataclass
class GPIBDeviceTransport(AbstractTransport):
    """Device specific transport, handles addressing transparently"""

    lltransport: GPIBTransport = field()
    address: Union[AddressTuple, int] = field()

    def __post_init__(self) -> None:
        """Make sure address is always tuple"""
        if isinstance(self.address, int):
            self.address = (self.address, None)

    async def _set_ll_address(self) -> None:
        """Set the lowlevel transport address"""
        assert not isinstance(self.address, int)
        await self.lltransport.set_address(self.address[0], self.address[1])

    async def send_command(self, command: str) -> None:
        """Transparently set address when talking, see low-level transport method docs for more info"""
        await self._set_ll_address()
        await self.lltransport.send_command(command)

    async def get_response(self) -> str:
        """Transparently set address when talking, see low-level transport method docs for more info"""
        await self._set_ll_address()
        return await self.lltransport.get_response()

    async def abort_command(self) -> None:
        """Transparently set address when talking, see low-level transport method docs for more info"""
        await self._set_ll_address()
        return await self.lltransport.abort_command()

    async def send_scd(self) -> None:
        """Sends the Selected Device Clear (SDC) message to the device"""
        await self._set_ll_address()
        await self.lltransport.send_scd()

    async def send_llo(self) -> None:
        """Send LLO (disable front panel) to the device"""
        await self._set_ll_address()
        await self.lltransport.send_llo()

    async def send_loc(self) -> None:
        """Send LOC (enable front panel) to the device"""
        await self._set_ll_address()
        await self.lltransport.send_loc()

    async def quit(self) -> None:
        """Nop for devices"""
        LOGGER.debug("quit should not be called on device level but we must have the method for type compatibility")


class GPIBTransport(BaseTransport, ABC):
    """Baseclass for GPIB transports"""

    @abstractmethod
    async def set_address(self, primary: int, secondary: Optional[int] = None) -> None:
        """Set the address we want to talk to"""
        raise NotImplementedError()

    @abstractmethod
    async def query_address(self) -> AddressTuple:
        """Query the address we are talking to, returns tuple with primary and secondary parts
        secondary is None if not set"""
        raise NotImplementedError()

    @abstractmethod
    async def scan_devices(self) -> Sequence[Tuple[int, str]]:
        """Scan for devices in the bus.
        Returns list of addresses and identifiers for found primary addresses (0-30)"""
        raise NotImplementedError()

    @abstractmethod
    async def send_scd(self) -> None:
        """Sends the Selected Device Clear (SDC) message to the currently specified GPIB address"""
        raise NotImplementedError()

    @abstractmethod
    async def send_ifc(self) -> None:
        """Asserts GPIB IFC signal"""
        raise NotImplementedError()

    @abstractmethod
    async def send_llo(self) -> None:
        """Send LLO (disable front panel) to currently specified address"""
        raise NotImplementedError()

    @abstractmethod
    async def send_loc(self) -> None:
        """Send LOC (enable front panel) to currently specified address"""
        raise NotImplementedError()

    @abstractmethod
    async def send_group_trig(self) -> None:
        """Send Group Execute Trigger to the bus"""
        raise NotImplementedError()

    @abstractmethod
    async def get_srq(self) -> int:
        """Get SRQ assertion status"""
        raise NotImplementedError()

    @abstractmethod
    async def poll(self) -> int:
        """Do serial poll on the selected device"""
        raise NotImplementedError()

    def get_device_transport(self, address: int, secondary: Optional[int] = None) -> GPIBDeviceTransport:
        """Gets a device-specific transport instance for given address"""
        return GPIBDeviceTransport(self, (address, secondary))



"""TDK Lambda power supplies"""
# pylint: disable=C0103
from typing import Optional, Union, Any
from dataclasses import dataclass, field
import decimal
import logging

import serial as pyserial  # type: ignore

from ..scpi import SCPIDevice, SCPIProtocol
from ..transports.rs232 import RS232Transport
from ..transports.tcp import get as get_tcp
from .generic import PowerSupply

StrIntCombo = Union[str, int]
LOGGER = logging.getLogger(__name__)


class TDKSCPI(SCPIDevice):
    """Baseclass for TDK SCPI devices"""

    async def set_power_on_status_clear(self, setting: StrIntCombo) -> None:
        """
        Set the Power-On Status Clear setting.
        * ON/1/True - This choice enables the power-on clearing of the listed registers
        * OFF/0/false - This choice disable the clearing of the listed registers and they retain
                        their status when a power-on condition occurs
        """
        setting = str(setting).upper()
        if setting in ("1", "ON", "TRUE"):
            setting = "1"
        elif setting in ("0", "OFF", "FALSE"):
            setting = "0"
        else:
            raise ValueError
        await super().set_power_on_status_clear(setting)

    async def restore_state(self, state: int) -> None:
        """
        Restores the power supply to a state previously stored in memory by *SAV command.
        """
        state = int(state)
        if state not in (1, 2, 3, 4):
            raise ValueError("invalid state")

        await super().restore_state(state)

    async def save_state(self, state: int) -> None:
        """
        The SAV command saves all applied configuration settings.
        """
        state = int(state)
        if state not in (1, 2, 3, 4):
            raise ValueError("Invalid state")

        await super().save_state(state)

    async def power_on_state(self, setting: StrIntCombo) -> None:
        """
        Set the power-on behavior of the system
        * 1 - AUTO - The power supply output will return to its previous value
                     when the latching fault condition is removed or to the
                     stored value after AC recycle.
        * 0 - SAFE - The power supply output will remain Off after the fault
                     condition is removed or after AC recycle.
        """
        setting = str(setting).upper()
        if setting in ("1", "ON", "TRUE"):
            setting = "1"
        elif setting in ("0", "OFF", "FALSE"):
            setting = "0"
        else:
            raise ValueError
        await super().power_on_state(setting)


@dataclass
class TDKLambdaZplus(PowerSupply, TDKSCPI):
    """TDK Lambda Z+ power supply"""

    voltage_limit: float = field(default=20.0)
    current_limit: float = field(default=10.0)

    async def measure_current(self, extra_params: str = "") -> decimal.Decimal:
        """
        Returns the actual output current in amps.

        extra_params: String to append to the command.  The only valid command
                for this device is ":DC"
        """

        resp = await self.ask(f"MEAS:CURR{extra_params}?")
        return decimal.Decimal(resp)

    async def measure_voltage(self, extra_params: str = "") -> decimal.Decimal:
        """
        Returns the actual output voltage in volts.

        extra_params: String to append to the command.  The only valid command
                for this device is ":DC"
        """

        resp = await self.ask(f"MEAS:VOLT{extra_params}?")
        return decimal.Decimal(resp)

    async def measure_power(self, extra_params: str = "") -> decimal.Decimal:
        """
        Returns the actual output power in watts.

        extra_params: String to append to the command.  The only valid command
                for this device is ":DC"
        """

        resp = await self.ask(f"MEAS:POW{extra_params}?")
        return decimal.Decimal(resp)

    async def select_active_instrument(self, select_id: int) -> None:
        """
        Select the power supply for communication.

        id: the ID of the power supply to select.  int from 1-31
        """

        _id = int(select_id)

        if _id < 1 or _id > 31:
            raise ValueError("id %d is outside of the valid id range" % _id)

        await self.command(f"INSTrument:NSELect {_id:d}")

    async def query_active_instrument(self) -> int:
        """
        Returns the ID of the active instrument.
        """

        resp = await self.ask("INSTrument:NSELect?")
        return int(resp)

    async def couple_mode(self, couple: str = "NONE") -> None:
        """
        Couple for all Z+ power supplies.
        """

        couple = couple.upper()

        if couple in ("NONE", "ALL"):
            await self.command("INSTrument:COUPle %s" % couple)
        else:
            raise ValueError("Argument '%s' not valid for INST:COUP" % couple)

    async def set_voltage_protection(self, volts: Any) -> None:
        """
        Set over-voltage protection level.
        """

        _volts = str(volts).upper()
        # FIXME: shouldn't we pass _volts here ?? Also what are valid types/values ??
        await self.command("VOLTage:PROTection:LEVel")

    async def query_voltage_protection(self, mode: Optional[str] = None) -> decimal.Decimal:
        """
        Query the voltage protection level.  Depending on mode, returns the current level, the
        minimum level, or the maximum level.

        mode: Which value to return.
                - None (default): returns the current voltage protection level
                - "MAX": returns the maximum possible voltage protection level
                - "MIN": returns the minimum possible voltage protection level, approx. 105% the
                         current voltage setting
        """

        if mode is None:
            resp = await self.ask("VOLTage:PROTection:LEVel?")
        else:
            resp = await self.ask(f"VOLTage:PROTection:LEVel? {mode}")
        return decimal.Decimal(resp)

    async def flash_display(self, setting: StrIntCombo) -> None:
        """
        Make the front panel voltage and Current displays flash.
        """

        setting = str(setting).upper()
        if setting in ("1", "ON", "TRUE"):
            setting = "1"
        elif setting in ("0", "OFF", "FALSE"):
            setting = "0"
        else:
            raise ValueError
        await self.command(f"DISPlay:FLASh {setting}")

    async def global_enable(self, setting: StrIntCombo) -> None:
        """
        Set enable status of all units.
        """

        setting = str(setting).upper()
        if setting in ("1", "ON", "TRUE"):
            setting = "1"
        elif setting in ("0", "OFF", "FALSE"):
            setting = "0"
        else:
            raise ValueError
        await self.command(f"GLOBal:OUTPut:STATe {setting}")

    async def global_set_voltage(self, volts: float) -> None:
        """
        Set enable status of all units.
        """

        _volts = str(volts)
        # FIXME: probably just using volts:f would work fine
        await self.command(f"GLOBal:VOLTage:AMPLitude {_volts}")

    async def global_reset(self) -> None:
        """
        Reset all units.
        """

        await self.command("GLOBal:*RST")

    async def set_voltage(self, millivolts: float, extra_params: str = "") -> None:
        """
        Sets the desired output voltage (but does not auto-enable outputs) in
        millivolts, pass extra_params string to append to the command (like ":PROT")

        Limited to five percent greater than the voltage limit of the unit.
        """

        if millivolts / 1000.0 > 1.05 * self.voltage_limit or millivolts < 0:
            raise ValueError

        await super().set_voltage(millivolts, extra_params=extra_params)

    async def set_current(self, milliamps: float, extra_params: str = "") -> None:
        """
        Sets the desired output current (but does not auto-enable outputs) in
        milliamps, pass extra_params string to append to the command (like ":TRIG")

        Limited to five percent greater than the current limit of the unit.
        """

        if milliamps / 1000.0 > 1.05 * self.current_limit or milliamps < 0:
            raise ValueError

        await super().set_current(milliamps, extra_params=extra_params)


def tcp(ipaddr: str, port: int) -> TDKLambdaZplus:
    """Quick helper to connect via TCP"""

    transport = get_tcp(ipaddr, port)
    protocol = SCPIProtocol(transport)
    dev = TDKLambdaZplus(protocol)
    return dev


def serial(serial_url: str, baudrate: int = 9600) -> TDKLambdaZplus:
    """ Quick helper to connect via serial """
    port = pyserial.serial_for_url(
        serial_url,
        baudrate=baudrate,
        bytesize=8,
        parity=pyserial.PARITY_NONE,
        stopbits=1,
        xonxoff=False,
        rtscts=False,
        dsrdtr=False,
        timeout=10,
    )
    transport = RS232Transport(serialdevice=port)
    protocol = SCPIProtocol(transport)
    dev = TDKLambdaZplus(protocol)
    return dev



#!/usr/bin/env python3
"""Example/test script for using the HP 6632B power supply via serial interface"""
import atexit
import os
import sys

from scpi.devices import hp6632b
from scpi.wrapper import AIOWrapper

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print(f"run with python -i {__file__} /dev/ttyUSB0")
        sys.exit(1)
    # Then put to interactive mode
    os.environ["PYTHONINSPECT"] = "1"
    aiodev = hp6632b.rs232(sys.argv[1], rtscts=True, baudrate=9600)
    dev = AIOWrapper(aiodev)

    atexit.register(dev.quit)

    print(dev.identify())



"""Baseclass for all the transports, if common methods are needed they will be defined here

All transports must define certain basic methods (check all the raise NotImplementedError)
"""
from typing import Optional, Callable
import asyncio
import logging
import threading
from abc import ABC, abstractmethod
from dataclasses import dataclass, field


LOGGER = logging.getLogger(__name__)


class AbstractTransport(ABC):  # pylint: disable=R0903
    """So that for example GPIBDeviceTransport can be identified as transport without inheriting
    the low-level transport methods"""


@dataclass  # type: ignore # there's something weird with dataclasses and ABCs
class BaseTransport(AbstractTransport, ABC):
    """Baseclass for SCPI tranport layers, abstracts away details, must be subclasses to implement"""

    message_callback: Optional[Callable[[str], None]] = field(default=None)
    unsolicited_message_callback: Optional[Callable[[str], None]] = field(default=None)
    lock: asyncio.Lock = field(default_factory=asyncio.Lock)
    aioevent: asyncio.Event = field(default_factory=asyncio.Event)
    blevent: threading.Event = field(default_factory=threading.Event)

    @abstractmethod
    async def quit(self) -> None:
        """Must shutdown all background threads (if any)"""
        raise NotImplementedError()

    @abstractmethod
    async def send_command(self, command: str) -> None:
        """Sends a complete command to the device, line termination, write timeouts etc are handled by the transport
        note: the transport probably should handle locking transparently using
        'with (await self.lock):' as context manager"""
        raise NotImplementedError()

    @abstractmethod
    async def get_response(self) -> str:
        """Tells the device send a response, reads and returns it"""
        raise NotImplementedError()

    def message_received(self, message: str) -> None:
        """Passes the message to the callback expecting it, or to the unsolicited callback"""
        if self.message_callback is not None:
            self.message_callback(message)
            self.message_callback = None
            return
        # Fall-through for unsolicited messages
        if self.unsolicited_message_callback is not None:
            self.unsolicited_message_callback(message)
            return
        LOGGER.info("Got unsolicited message but have no callback to send it to")

    @abstractmethod
    async def abort_command(self) -> None:
        """Send the "device clear" command to abort a running command
        note: the transport probably should handle locking transparently using
        'async with self.lock:' as context manager"""
        raise NotImplementedError()



""""Driver" for http://prologix.biz/gpib-usb-controller.html GPIB controller"""
from typing import Optional, Sequence, List, Any, Tuple, cast
import asyncio
import logging
from dataclasses import dataclass

import serial  # type: ignore
import serial.threaded  # type: ignore

from ..rs232 import RS232SerialProtocol, RS232Transport
from .base import GPIBTransport, AddressTuple

SCAN_DEVICE_TIMEOUT = 0.5
READ_TIMEOUT = 1.0
LOGGER = logging.getLogger(__name__)


class PrologixRS232SerialProtocol(RS232SerialProtocol):
    """Basically the same deal as with the stock RS232 PySerial "protocol" but different EOL"""

    TERMINATOR = b"\n"


@dataclass
class PrologixGPIBTransport(GPIBTransport, RS232Transport):
    """Transport "driver" for the Prologix USB-GPIB controller (v6 protocol)"""

    def __post_init__(self) -> None:
        """Init the serial and controller"""
        self._serialhandler = serial.threaded.ReaderThread(self.serialdevice, PrologixRS232SerialProtocol)
        self._serialhandler.start()
        self._serialhandler.protocol.handle_line = self.message_received
        self.initialize_controller()

    def initialize_controller(self) -> None:
        """Initializes the controller to known state"""
        if not self._serialhandler:
            raise RuntimeError("Serialhandler isn't")
        # Set to controller mode
        self._serialhandler.protocol.write_line("++mode 1")
        # Disable automatic read after write
        self._serialhandler.protocol.write_line("++auto 0")
        # Auto-assert End Of Instruction after commands
        self._serialhandler.protocol.write_line("++eoi 1")
        # Append CRLF to device commands (EOI above *should* be enough but this is probably more compatible)
        self._serialhandler.protocol.write_line("++eos 0")
        # We do not have parsing support for the EOT character so disable it
        self._serialhandler.protocol.write_line("++eot_enable 0")
        # Set inter-character timeout for read commands
        self._serialhandler.protocol.write_line("++read_tmo_ms 500")
        # Assert IFC, make us Controller In Charge
        self._serialhandler.protocol.write_line("++ifc")

    async def send_command(self, command: str) -> None:
        """Wrapper for write_line on the protocol with some sanity checks"""
        if not self._serialhandler or not self._serialhandler.is_alive():
            raise RuntimeError("Serial handler not ready")
        async with self.lock:
            self._serialhandler.protocol.write_line(command)

    async def get_response(self) -> str:
        """Get device response"""
        return await self.send_and_read("++read eoi")

    async def send_and_read(self, send: str) -> str:
        """Send a line, read the response. NOTE: This is for talking with the controller, device responses
        need to use get_response as usual"""
        if not self._serialhandler:
            raise RuntimeError("Serialhandler isn't")

        async def _send_and_read(send: str) -> str:
            """Wrap the actual work"""
            nonlocal self
            if not self._serialhandler:
                raise RuntimeError("Serialhandler isn't")

            async with self.lock:
                response: Optional[str] = None

                def set_response(message: str) -> None:
                    """Callback for setting the response"""
                    nonlocal response, self
                    response = message
                    self.blevent.set()

                self.blevent.clear()
                self.message_callback = set_response
                self._serialhandler.protocol.write_line(send)
                await asyncio.get_event_loop().run_in_executor(None, self.blevent.wait)
                self.message_callback = None
                return cast(str, response)

        return await asyncio.wait_for(_send_and_read(send), timeout=READ_TIMEOUT)

    async def set_address(self, primary: int, secondary: Optional[int] = None) -> None:
        """Set the address we want to talk to"""
        if secondary is None:
            await self.send_command(f"++addr {primary:d}")
        else:
            await self.send_command(f"++addr {primary:d} {secondary:d}")

        while True:
            await asyncio.sleep(0.001)
            resp = await self.query_address()
            if resp == (primary, secondary):
                break

    async def query_address(self) -> AddressTuple:
        """Query the address we are talking to, returns tuple with primary and secondary parts
        secondary is None if not set"""
        resp = await self.send_and_read("++addr")
        parts = resp.split(" ")
        primary = int(parts[0])
        secondary: Optional[int] = None
        if len(parts) > 1:
            secondary = int(parts[1])
        return (primary, secondary)

    async def send_scd(self) -> None:
        """Sends the Selected Device Clear (SDC) message to the currently specified GPIB address"""
        await self.send_command("++clr")

    async def send_ifc(self) -> None:
        """Asserts GPIB IFC signal"""
        await self.send_command("++ifc")

    async def send_llo(self) -> None:
        """Send LLO (disable front panel) to currently specified address"""
        await self.send_command("++llo")

    async def send_loc(self) -> None:
        """Send LOC (enable front panel) to currently specified address"""
        await self.send_command("++loc")

    async def get_srq(self) -> int:
        """Get SRQ assertion status"""
        resp = await self.send_and_read("++srq")
        return int(resp)

    async def poll(self) -> int:
        """Do serial poll on the selected device"""
        resp = await self.send_and_read("++spoll")
        return int(resp)

    async def send_group_trig(self, addresses: Optional[Sequence[int]] = None) -> None:  # pylint: disable=W0221
        """Send trigger to listed addresses

        For some reason Prologix does not trigger the whole bus but only listed devices (if none listed then
        the currently selected device is used)"""
        if addresses is None:
            return await self.send_command("++trg")
        await self.send_command("++trg " + " ".join((str(x) for x in addresses)))

    async def scan_devices(self) -> Sequence[Tuple[int, str]]:
        """Scan for devices in the bus.
        Returns list of addresses and identifiers for found primary addresses (0-30)"""
        if not self._serialhandler:
            raise RuntimeError("Serialhandler isn't")
        found_addresses: List[int] = []
        # We do not lock on this level since the commands we use need to manipulate the lock
        prev_addr = await self.query_address()
        prev_read_tmo_ms = int(await self.send_and_read("++read_tmo_ms"))
        new_read_tmo_ms = int((SCAN_DEVICE_TIMEOUT / 2) * 1000)
        self._serialhandler.protocol.write_line(f"++read_tmo_ms {new_read_tmo_ms:d}")
        for addr in range(0, 31):  # 0-30 inclusive

            async def _scan_addr(addr: int) -> None:
                """Sacn single address"""
                nonlocal found_addresses, self
                await self.set_address(addr)
                await self.poll()
                found_addresses.append(addr)

            try:
                await asyncio.wait_for(_scan_addr(addr), timeout=SCAN_DEVICE_TIMEOUT)
            except (asyncio.TimeoutError, asyncio.CancelledError):
                pass
        self._serialhandler.protocol.write_line(f"++read_tmo_ms {prev_read_tmo_ms:d}")
        # Wait a moment for things to settle
        await asyncio.sleep(float(prev_read_tmo_ms) / 1000)
        # Get ids for the devices we found
        ret = []
        for addr in found_addresses:
            await self.set_address(addr)
            await self.send_command("*IDN?")
            idstr = await self.get_response()
            ret.append((addr, idstr))
        await self.set_address(*prev_addr)
        return ret

    async def abort_command(self) -> None:
        """Not implemented for prologix"""
        LOGGER.debug("not implemented on PrologixGPIBTransport")


def get(serial_url: str, **serial_kwargs: Any) -> PrologixGPIBTransport:
    """Shorthand for creating the port from url and initializing the transport"""
    port = serial.serial_for_url(serial_url, **serial_kwargs)
    return PrologixGPIBTransport(serialdevice=port)



"""
Created on febrary 21 2020.

@author: qmor
"""

import atexit
import os
import time

from scpi.devices import TDKLambdaZPlus
from scpi.wrapper import AIOWrapper

if __name__ == "__main__":
    # Then put to interactive mode
    os.environ["PYTHONINSPECT"] = "1"
    aiodev = TDKLambdaZPlus.tcp("192.168.3.34", 8003)
    dev = AIOWrapper(aiodev)
    atexit.register(dev.quit)
    dev.identify()
    while True:
        dev.query_voltage()
        dev.query_current()
        dev.query_output()
        time.sleep(1)



"""Generic SCPI commands, allow sending and reading of raw data, helpers to parse information"""
from typing import Any, Tuple, Sequence, Union, Optional, cast
import asyncio
import re
import logging
from dataclasses import dataclass, field

from .errors import CommandError
from .transports.baseclass import AbstractTransport, BaseTransport
from .transports.gpib import GPIBDeviceTransport, GPIBTransport


COMMAND_DEFAULT_TIMEOUT = 1.0
ERROR_RE = re.compile(r'([+-]?\d+),"(.*?)"')
LOGGER = logging.getLogger(__name__)

# FIXME rename to mixin and use as such
class BitEnum:  # pylint: disable=R0903
    """Baseclass for bit definitions of various status registers"""

    @classmethod
    def test_bit(cls, statusvalue: int, bitname: str) -> bool:
        """Test if the given status value has the given bit set"""
        bitval = cast(int, getattr(cls, bitname))
        return bool(bitval & statusvalue)


# FIXME use enum.IntEnum as baseclass (maybe, we can't do the docstrings the names if we do...)
class ESRBit(BitEnum):
    """Define meanings of the Event Status Register (ESR) bits"""

    @property
    def power_on(self) -> int:
        """Power-on. The power has cycled"""
        return 128

    @property
    def user_request(self) -> int:
        """User request. The instrument operator has issued a request,
        for instance turning a knob on the front panel."""
        return 64

    @property
    def command_error(self) -> int:
        """Command Error. A command error has occurred."""
        return 32

    @property
    def exec_error(self) -> int:
        """Execution error. The instrument was not able to execute a command for
        some reason. The reason can be that the supplied data is out of range but
        can also be an external event like a safety switch/knob or some hardware /
        software error."""
        return 16

    @property
    def device_error(self) -> int:
        """Device Specific Error."""
        return 8

    @property
    def query_error(self) -> int:
        """Query Error. Error occurred during query processing."""
        return 4

    @property
    def control_request(self) -> int:
        """Request Control. The instrument is requesting to become active controller."""
        return 2

    @property
    def operation_complete(self) -> int:
        """Operation Complete. The instrument has completed all operations.
        This bit is used for synchronisation purposes."""
        return 1


class STBBit(BitEnum):
    """Define meanings of the STatus Byte register (STB) bits"""

    @property
    def rqs_mss(self) -> int:
        """RQS, ReQuested Service. This bit is set when the instrument has requested
        service by means of the SeRvice Request (SRQ). When the controller reacts
        by performing a serial poll, the STatus Byte register (STB) is transmitted with
        this bit set. Afand cleared afterwards. It is only set again when a new event
        occurs that requires service.

        MSS, Master Summary Status. This bit is a summary of the STB and the
        SRE register bits 1..5 and 7. Thus it is not cleared when a serial poll occurs.
        It is cleared when the event which caused the setting of MSS is cleared or
        when the corresponding bits in the SRE register are cleared."""
        return 64

    @property
    def rqs(self) -> int:
        """alias for rqs_mss"""
        return self.rqs_mss

    @property
    def mss(self) -> int:
        """alias for rqs_mss"""
        return self.rqs_mss

    @property
    def esb(self) -> int:
        """ESB, Event Summary Bit. This is a summary bit of the standard status
        registers ESR and ESE"""
        return 32

    @property
    def event_summary(self) -> int:
        """Alias for esb"""
        return self.esb

    @property
    def mav(self) -> int:
        """MAV, Message AVailable. This bit is set when there is data in the output
        queue waiting to be read."""
        return 16

    @property
    def message_available(self) -> int:
        """Alias for mav"""
        return self.mav

    @property
    def eav(self) -> int:
        """EAV, Error AVailable. This bit is set when there is data in the output
        queue waiting to be read."""
        return 4

    @property
    def error_available(self) -> int:
        """Alias for eav"""
        return self.eav


@dataclass
class SCPIProtocol:
    """Implements the SCPI protocol talks over the given transport"""

    transport: Union[BaseTransport, GPIBDeviceTransport, GPIBTransport] = field()
    lock: asyncio.Lock = field(default_factory=asyncio.Lock)
    _checking_error: bool = field(default=False)

    async def quit(self) -> None:
        """Shuts down any background threads that might be active"""
        await self.transport.quit()

    async def abort_command(self) -> None:
        """Shortcut to the transports abort_command call"""
        await self.transport.abort_command()

    async def get_error(self) -> Tuple[int, str]:
        """Asks for the error code and string"""
        if self._checking_error:
            raise RuntimeError("Recursion on get_error detected")
        try:
            self._checking_error = True
            response = await self.ask("SYST:ERR?", auto_check_error=False)
            match = ERROR_RE.search(response)
            if not match:
                # PONDER: Make our own exceptions ??
                raise ValueError("Response '{:s}' does not have correct error format".format(response))
            code = int(match.group(1))
            errstr = match.group(2)
            return (code, errstr)
        finally:
            self._checking_error = False

    async def check_error(self, prev_command: str = "") -> None:
        """Check for error and raise exception if present"""
        code, errstr = await self.get_error()
        if code != 0:
            raise CommandError(prev_command, code, errstr)

    async def command(
        self,
        command: str,
        cmd_timeout: float = COMMAND_DEFAULT_TIMEOUT,
        abort_on_timeout: bool = True,
        *,
        auto_check_error: bool = True,
    ) -> None:
        """Sends a command, does not wait for response"""
        try:

            async def _command(command: str) -> None:
                """Wrap the actual work"""
                nonlocal self
                async with self.lock:
                    await self.transport.send_command(command)

            await asyncio.wait_for(_command(command), timeout=cmd_timeout)
        except asyncio.TimeoutError as err:
            # check for the actual error if available
            if auto_check_error:
                await self.check_error(command)
            if abort_on_timeout:
                await self.abort_command()
            # re-raise the timeout if no other error found
            raise err
        except asyncio.CancelledError:
            LOGGER.info("Cancelled")
        # other errors are allowed to bubble-up as-is

    async def safe_command(self, command: str, *args: Any, **kwargs: Any) -> None:
        """See "command", this just auto-checks for errors each time"""
        await self.command(command, *args, **kwargs)
        await self.check_error(command)

    async def ask(
        self,
        command: str,
        cmd_timeout: float = COMMAND_DEFAULT_TIMEOUT,
        abort_on_timeout: bool = True,
        *,
        auto_check_error: bool = True,
    ) -> str:
        """Send a command and waits for response, returns the response"""
        try:

            async def _ask(command: str) -> str:
                """Wrap the actual work"""
                nonlocal self
                async with self.lock:
                    await self.transport.send_command(command)
                    return await self.transport.get_response()

            return await asyncio.wait_for(_ask(command), timeout=cmd_timeout)
        except asyncio.TimeoutError as err:
            # check for the actual error if available
            if auto_check_error:
                await self.check_error(command)
            if abort_on_timeout:
                self.abort_command()
            # re-raise the timeout if no other error found
            raise err
        except asyncio.CancelledError:
            LOGGER.info("Cancelled")
            # gotta return something or raise an error
            raise
        # other errors are allowed to bubble-up as-is

    async def safe_ask(self, command: str, *args: Any, **kwargs: Any) -> str:
        """See "ask", this just autp-checks for errors each time"""
        response = await self.ask(command, *args, **kwargs)
        await self.check_error(command)
        return response


@dataclass
class SCPIDevice:  # pylint: disable=R0904
    """Implements nicer wrapper methods for the raw commands from the generic SCPI command set

    See also devices.mixins for mixin classes with more features"""

    instancefrom: Union[BaseTransport, "SCPIDevice", SCPIProtocol, GPIBDeviceTransport, GPIBTransport]
    use_safe_variants: bool = field(default=True)
    protocol: SCPIProtocol = field(init=False)
    transport: AbstractTransport = field(init=False)
    _can_poll: bool = field(default=False)

    def __post_init__(self) -> None:
        """Set protocol and transport based on what we're instancing from"""
        protocol: Optional[SCPIProtocol] = None
        if isinstance(self.instancefrom, SCPIProtocol):
            protocol = self.instancefrom
        if isinstance(self.instancefrom, (BaseTransport, GPIBDeviceTransport, GPIBTransport)):
            protocol = SCPIProtocol(self.instancefrom)
        if isinstance(self.instancefrom, SCPIDevice):
            protocol = self.instancefrom.protocol
        if not protocol:
            raise RuntimeError("Could not resolve protocol/transport")
        self.protocol = protocol
        self.transport = self.protocol.transport
        # Check if transport poll method exists
        # TODO: the transport class should have a marker property for this we should use
        try:
            _ = self.transport.poll  # type: ignore
            self._can_poll = True
        except AttributeError:
            pass

    async def command(
        self, command: str, cmd_timeout: float = COMMAND_DEFAULT_TIMEOUT, abort_on_timeout: bool = True
    ) -> None:
        """Wrap the protocol command (using safe version if requested)"""
        if self.use_safe_variants:
            return await self.protocol.safe_command(command, cmd_timeout, abort_on_timeout)
        return await self.protocol.command(command, cmd_timeout, abort_on_timeout)

    async def ask(
        self, command: str, cmd_timeout: float = COMMAND_DEFAULT_TIMEOUT, abort_on_timeout: bool = True
    ) -> str:
        """Wrap the protocol ask (using safe version if requested)"""
        if self.use_safe_variants:
            return await self.protocol.safe_ask(command, cmd_timeout, abort_on_timeout)
        return await self.protocol.ask(command, cmd_timeout, abort_on_timeout)

    async def quit(self) -> None:
        """Shuts down any background threads that might be active"""
        await self.protocol.quit()

    async def abort(self) -> None:
        """Tells the protocol layer to issue "Device clear" to abort the command currently hanging"""
        await self.protocol.abort_command()

    async def get_error(self) -> Tuple[int, str]:
        """Shorthand for procotols method of the same name"""
        return await self.protocol.get_error()

    async def reset(self) -> None:
        """Resets the device to known state (with *RST) and clears the error log"""
        return await self.protocol.command("*RST;*CLS")

    async def wait_for_complete(self, wait_timeout: float) -> bool:
        """Wait for all queued operations to complete (up-to defined timeout)"""
        resp = await self.ask("*WAI;*OPC?", cmd_timeout=wait_timeout)
        return bool(int(resp))

    async def identify(self) -> Sequence[str]:
        """Returns the identification data, standard order is:
        Manufacturer, Model no, Serial no (or 0), Firmware version"""
        resp = await self.ask("*IDN?")
        return resp.split(",")

    async def query_esr(self) -> int:
        """Queries the event status register (ESR) NOTE: The register is cleared when read!
        returns int instead of Decimal like the other number queries since we need to be able
        to do bitwise comparisons"""
        resp = await self.ask("*ESR?")
        return int(resp)

    async def query_ese(self) -> int:
        """Queries the event status enable (ESE).
        returns int instead of Decimal like the other number queries since we need to be able
        to do bitwise comparisons"""
        resp = await self.ask("*ESE?")
        return int(resp)

    async def set_ese(self, state: int) -> None:
        """Sets ESE to given value.
        Construct the value with bitwise OR operations using ESRBit properties, for example to enable OPC and exec_error
        error bits in the status flag use: set_ese(ESRBit.operation_complete | ESRBit.exec_error)"""
        await self.command(f"*ESE {state:d}")

    async def query_sre(self) -> int:
        """Queries the service request enable (SRE).
        returns int instead of Decimal like the other number queries since we need to be able
        to do bitwise comparisons"""
        resp = await self.ask("*SRE?")
        return int(resp)

    async def set_sre(self, state: int) -> None:
        """Sets SRE to given value.
        Construct the value with bitwise OR operations using STBBit properties, for example to enable SRQ generation
        on any error or message  use: set_sre(STBBit.mav | STBBit.eav)"""
        await self.command(f"*SRE {state:d}")

    async def query_stb(self) -> int:
        """Queries the status byte (STB).
        returns int instead of Decimal like the other number queries since we need to be able
        to do bitwise comparisons

        If transport implements "serial poll", will use that instead of SCPI query to get the value"""
        if self._can_poll:
            resp = await self.transport.poll()  # type: ignore
        else:
            resp = await self.ask("*STB?")
        return int(resp)

    async def trigger(self) -> None:
        """Send the TRiGger command via SCPI.
        NOTE: For GPIB devices the Group Execute Trigger is way better, use it when possible
        however we do not do it transparently here since it triggers all devices on the bus"""
        await self.command("*TRG")

    async def clear_status(self) -> None:
        """
        Sends a clear status command.
        """
        await self.command("*CLS")

    async def operation_complete(self) -> None:
        """
        Sends an Operation Complete command.
        """
        await self.command("*OPC")

    async def query_options(self) -> str:
        """
        Queries the model's options.
        """
        return await self.ask("*OPT?")

    async def set_power_on_status_clear(self, setting: str) -> None:
        """
        Set the Power-On Status Clear setting.
        """
        await self.command(f"*PSC {setting}")

    async def save_state(self, state: int) -> None:
        """
        The SAV command saves all applied configuration settings.
        """
        state = int(state)
        await self.command("*SAV {state:d}")

    async def restore_state(self, state: int) -> None:
        """
        Restores the power supply to a state previously stored in memory by *SAV command.
        """
        state = int(state)
        await self.command(f"*RCL {state:d}")

    async def power_on_state(self, setting: str) -> None:
        """
        Set the power-on behavior of the system
        """
        setting = str(setting).upper()
        await self.command(f"*OUTP:PON {setting}")



"""SCPI module specific errors"""
from typing import Any


class CommandError(RuntimeError):
    """Error executing SCPI command"""

    def __init__(self, command: str, code: int, message: str) -> None:
        """initialize the error"""
        self.command = command
        self.code = code
        self.message = message
        super().__init__()

    def __str__(self) -> str:
        """format as string"""
        return f"'{self.command}' returned error {self.code:d}: {self.message}"



"""SCPI module, the scpi class implements the base command set, devices may extend it.
transports are separate from devices (so you can use for example hp6632b with either serial port or GPIB)"""
__version__ = "2.4.0"  # NOTE Use `bump2version --config-file patch` to bump versions correctly
from .scpi import SCPIProtocol, SCPIDevice
from .errors import CommandError

__all__ = ["SCPIProtocol", "CommandError", "SCPIDevice"]



"""Helper class to allow using of device in traditional blocking style without having to deal with the ioloop"""
from typing import Any
import asyncio
import functools
import inspect
import logging


LOGGER = logging.getLogger(__name__)


class AIOWrapper:  # pylint: disable=R0903
    """Wraps all coroutine methods into asyncio run_until_complete calls"""

    def __init__(self, to_be_wrapped: Any) -> None:
        """Init wrapper for device"""
        self._device = to_be_wrapped
        self._loop = asyncio.get_event_loop()
        for attr in functools.WRAPPER_ASSIGNMENTS:
            try:
                setattr(self, attr, getattr(self._device, attr))
            except AttributeError:
                try:
                    setattr(self.__class__, attr, getattr(self._device.__class__, attr))
                except AttributeError:
                    LOGGER.debug("Could not copy {}".format(attr))

    def __getattr__(self, item: str) -> Any:
        """Get a memeber, if it's a coroutine autowrap it to eventloop run"""
        orig = getattr(self._device, item)
        if inspect.iscoroutinefunction(orig):

            @functools.wraps(orig)
            def wrapped(*args: Any, **kwargs: Any) -> Any:
                """Gets the waitable and tells the event loop to run it"""
                nonlocal self
                waitable = orig(*args, **kwargs)
                return self._loop.run_until_complete(waitable)

            return wrapped
        return orig

    def __dir__(self) -> Any:
        """Proxy the dir on the device"""
        return dir(self._device)

    def quit(self) -> None:
        """Calls the device.quit via loop and closes the loop"""
        self._loop.run_until_complete(self._device.quit())
        self._loop.close()


class DeviceWrapper(AIOWrapper):  # pylint: disable=R0903
    """Legacy name for the AsyncIO wrapper class for backwards compatibility"""
